#!/usr/bin/env python3

import socket
import sys
import random
import string
from multiprocessing import Process, Manager

manager = Manager()

class Recvlines:
    def __init__(self, sock):
        self.data = manager.dict()
        self.data["sock"] = sock
        self.data["buf"] = bytearray()
        self.data["done"] = False

    def __iter__(self):
        return self

    def __next__(self):
        if self.data["done"]:
            raise StopIteration

        buf = self.data["buf"]
        sock = self.data["sock"]

        while b"\n" not in buf:
            new_data = sock.recv(10_000_000)
            if not new_data:
                self.data["done"] = True
                if buf:
                    return buf
                else:
                    raise StopIteration
            buf.extend(new_data)

        line, buf = buf.split(b"\n", 1)

        self.data["sock"] = sock
        self.data["buf"] = buf

        return line + b"\n"

def make_listening_sock(addr, port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind((addr, port))
    sock.listen()
    return sock

def report_ready(control_conn):
    control_conn.sendall(b"+")

def report_ok(control_conn):
    control_conn.sendall(b"0")

def report_fail(control_conn):
    control_conn.sendall(b"1")

def recvall(sock):
    received = b""
    while True:
        new_data = sock.recv(65536)
        if not new_data:
            break
        received += new_data
    return received

def check_message(line_iter, expected):
    actual_message = next(line_iter)
    if actual_message == expected:
        sys.exit(0)
    else:
        print(f"Wrong message: {actual_message} instead of {expected}")
        sys.exit(1)

def check_messages_ok(line_iters, expected):
    processes = []
    for line_iter in line_iters:
        processes.append(Process(target=check_message, args=(line_iter, expected)))

    for process in processes:
        process.start()

    failed = False

    for process in processes:
        process.join()
        if process.exitcode != 0:
            failed = True

    return not failed

def driver_echo(control_conn):
    with make_listening_sock("localhost", 8091) as test_sock:
        report_ready(control_conn)

        with test_sock.accept()[0] as conn:
            received = b""
            while True:
                new_data = conn.recv(65536)
                if not new_data:
                    break
                received += new_data
            conn.sendall(received)

    control_conn.sendall(b"0")

def one_message_one_addr(control_conn):
    with make_listening_sock("localhost", 8091) as test_sock:
        report_ready(control_conn)

        with test_sock.accept()[0] as conn:
            message = recvall(conn)

        if message == b"Lorem ipsum dolor sit amet\n":
            report_ok(control_conn)
        else:
            print(f"Wrong message: {message}")
            report_fail(control_conn)

def one_message_many_addrs(control_conn):
    test_socks = [make_listening_sock("localhost", port) for port in range(8091, 8096)]
    report_ready(control_conn)

    test_conns = [sock.accept()[0] for sock in test_socks]

    messages = [recvall(conn) for conn in test_conns]

    if all(msg == b"Lorem ipsum dolor sit amet\n" for msg in messages):
        report_ok(control_conn)
    else:
        print(f"Wrong messages: {messages}")
        report_fail(control_conn)

    for conn in test_conns:
        conn.close()
    for sock in test_socks:
        sock.close()

def many_messages_one_addr(control_conn):
    with make_listening_sock("localhost", 8091) as test_sock:
        report_ready(control_conn)

        with test_sock.accept()[0] as conn:
            messages = list(Recvlines(conn))

        if messages == [b"Lorem\n", b"ipsum\n", b"dolor\n", b"sit\n", b"amet\n"]:
            report_ok(control_conn)
        else:
            print(f"Wrong messages: {messages}")
            report_fail(control_conn)

def many_messages_many_addrs_generic(control_conn, messages):
    test_socks = [make_listening_sock("localhost", port) for port in range(8091, 8096)]
    report_ready(control_conn)

    test_conns = [sock.accept()[0] for sock in test_socks]

    failed = False

    line_iters = [Recvlines(conn) for conn in test_conns]

    for expected_message in messages:
        if not check_messages_ok(line_iters, expected_message):
            failed = True

    if failed:
        report_fail(control_conn)
    else:
        report_ok(control_conn)

    for conn in test_conns:
        conn.close()
    for sock in test_socks:
        sock.close()

def many_messages_many_addrs(control_conn):
    many_messages_many_addrs_generic(control_conn, [b"Lorem\n", b"ipsum\n", b"dolor\n", b"sit\n", b"amet\n"])

def make_random_lines(line_size):
    random.seed(10)
    lines = list(
        bytes(
            "".join(random.choices(string.ascii_lowercase, k=line_size)),
            encoding='ascii',
        )
        for _ in range(5)
    )
    return [line + b"\n" for line in lines]

def long_messages(control_conn):
    many_messages_many_addrs_generic(control_conn, make_random_lines(10_000_000))

def sequential(control_conn):
    test_socks = [make_listening_sock("localhost", port) for port in range(8091, 8096)]
    report_ready(control_conn)

    test_conns = [sock.accept()[0] for sock in test_socks]

    failed = False

    line_iters = [Recvlines(conn) for conn in test_conns]

    for expected_message in make_random_lines(100_000):
        if not check_messages_ok(line_iters, expected_message):
            failed = True

        # Acknowledge message
        control_conn.sendall(b">")

    if failed:
        report_fail(control_conn)
    else:
        report_ok(control_conn)

    for conn in test_conns:
        conn.close()
    for sock in test_socks:
        sock.close()

def send_error(control_conn):
    test_socks = [make_listening_sock("localhost", port) for port in range(8091, 8096)]
    report_ready(control_conn)

    test_conns = [sock.accept()[0] for sock in test_socks]

    failed = False

    line_iters = [Recvlines(conn) for conn in test_conns]

    expected_messages = make_random_lines(1_000_000)

    if not check_messages_ok(line_iters, expected_messages[0]):
        failed = True

    for conn in test_conns[:3]:
        conn.shutdown(socket.SHUT_RDWR)
        conn.close()
    for sock in test_socks[:3]:
        sock.close()

    # Acknowledge first message
    control_conn.sendall(b">")

    for expected_message in expected_messages[1:]:
        if not check_messages_ok(line_iters[3:], expected_message):
            failed = True

    if failed:
        report_fail(control_conn)
    else:
        report_ok(control_conn)

    for conn in test_conns[3:]:
        conn.close()
    for sock in test_socks[3:]:
        sock.close()

def reconnect(control_conn):
    test_socks = [make_listening_sock("localhost", port) for port in range(8091, 8096)]
    report_ready(control_conn)

    test_conns = [sock.accept()[0] for sock in test_socks]

    failed = False

    line_iters = [Recvlines(conn) for conn in test_conns]

    expected_messages = make_random_lines(1_000_000)

    # ===== Message 1 =====

    if not check_messages_ok(line_iters, expected_messages[0]):
        failed = True

    assert control_conn.recv(1) == b">", "invalid control command from client"

    for conn in test_conns[:4]:
        conn.shutdown(socket.SHUT_RDWR)
        conn.close()
    for sock in test_socks[:2]:
        sock.close()

    control_conn.sendall(b">")

    # ===== Message 2 =====

    if not check_messages_ok(line_iters[4:], expected_messages[1]):
        failed = True

    assert control_conn.recv(1) == b">", "invalid control command from client"
    control_conn.sendall(b">")

    # ===== Message 3 =====

    test_conns[2:4] = [sock.accept()[0] for sock in test_socks[2:4]]
    line_iters[2:4] = [Recvlines(conn) for conn in test_conns[2:4]]

    if not check_messages_ok(line_iters[2:], expected_messages[2]):
        failed = True

    assert control_conn.recv(1) == b">", "invalid control command from client"

    test_socks[1] = make_listening_sock("localhost", 8092)
    control_conn.sendall(b">")

    # ===== Message 4 =====

    test_conns[1] = test_socks[1].accept()[0]
    line_iters[1] = Recvlines(test_conns[1])

    if not check_messages_ok(line_iters[1:], expected_messages[3]):
        failed = True

    assert control_conn.recv(1) == b">", "invalid control command from client"

    test_socks[0] = make_listening_sock("localhost", 8091)

    control_conn.sendall(b">")

    # ===== Message 5 =====

    test_conns[0] = test_socks[0].accept()[0]
    line_iters[0] = Recvlines(test_conns[0])

    if not check_messages_ok(line_iters, expected_messages[4]):
        failed = True

    assert control_conn.recv(1) == b">", "invalid control command from client"
    control_conn.sendall(b">")

    if failed:
        report_fail(control_conn)
    else:
        report_ok(control_conn)

    for conn in test_conns[:4]:
        conn.close()
    for sock in test_socks[:2]:
        sock.close()

TESTS = {
    "A": driver_echo,
    "B": one_message_one_addr,
    "C": one_message_many_addrs,
    "D": many_messages_one_addr,
    "E": many_messages_many_addrs,
    "F": long_messages,
    "G": sequential,
    "H": send_error,
    "I": reconnect,
}

def handle_client(control_conn):
    while True:
        test_id = control_conn.recv(1)
        if not test_id:
            break
        test_id = test_id.decode('ascii')

        print(f"Enter test {test_id}", file=sys.stderr)
        TESTS[test_id](control_conn)
        print(f"Exit test {test_id}", file=sys.stderr)

def main():
    with make_listening_sock("localhost", 8090) as control_sock:
        while True:
            with control_sock.accept()[0] as control_conn:
                handle_client(control_conn)

if __name__ == '__main__':
    main()
